# PET-MAD
PET_MAD_LATEST_STABLE_VERSION = "1.0.2"
PET_MAD_UQ_AVAILABILITY_VERSION = "1.0.2"
PET_MAD_NC_AVAILABILITY_VERSION = "1.1.0"
PET_MAD_AVAILABLE_VERSIONS = ["1.1.0", "1.0.2", "1.0.1", "1.0.0"]
# PET-MAD DOS
PET_MAD_DOS_LATEST_STABLE_VERSION = "1.0"
PET_MAD_DOS_AVAILABLE_VERSIONS = ["1.0"]
